"""
This is a sample class for setting up, running and output of a model.
In this instance a shallow ice sheet model.
"""

# standard libraries
import sys
import os
import copy
import commands

# numerics
import numpy as N
import pylab as P
## this is the old numerics module, outdate, but we need it because of
## the netCDF libaray we use
import Numeric

# our modules
import helper_module
sys.path.append('fortran')
import kees_simple_ice_model


class SIA_model(object):
    """
    The class doing all the model setup, running, output and some
    plotting.
    """
    def __init__(self, conf_or_netcdf_file):
        """The standard initialisation function.

        @type conf_or_netcdf_file: string
        @param conf_or_netcdf_file: A config file or a former netCDF
                                    output file.
        """
        self.input_file = conf_or_netcdf_file

        #: dict holding the L{Dimension} instances
        self.dimensions = {}
        #: dict holding the L{Variable} instances
        self.variables = {}
        #: dict holding the global attributes
        self.global_attrs = {}
        self.global_attrs['svn_revison'] = self.get_svn_verison_number()

        if conf_or_netcdf_file[-3:]=='.nc':
            # if a .nc file read the netCDF file
            self.read_netCDF()
        elif conf_or_netcdf_file[-4:]=='.CDL':
            # if a .CDL file read the CDL file
            self.read_model_config()
            self.read_model_input()



    def read_model_config(self):
        """
        Read the configuration file of the model.
        """
        tmp_file = 'tmp/config.nc'
        # first remove old files from the tmp folder
        os.system('rm '+ tmp_file)
        # turn the CDL file into a .nc file
        helper_module.CDL2nc(self.input_file, tmp_file)
        # read it
        config_file = helper_module.open_netCDF_into_python(tmp_file)

        ## now comes the hard bit: fill the class with
        ## the specified data

        # global attrs: just set them to what they are in the config
        # file
        self.global_attrs.update(self.fix_Scientific_bug(config_file.__dict__))

        # read the input files and put their data into the appropriate
        # place
        for key in self.global_attrs:
            ## make input variables
            if key.startswith('input_file'):
                input_file = helper_module.open_netCDF_into_python(
                    self.global_attrs[key])
                # here we assume that the dimension (X) is the same for all
                # input and of the same length.  Thus set the dimension
                # to that length:
                if self.dimensions=={}:
                    self.createDimension(
                        'X', input_file.dimensions['X'])
                    # and the corresponding coordinate variable
                    X = input_file.variables['X']
                    # make into numpy type, see note below:
                    dict_ = self.fix_Scientific_bug(X.__dict__)
                    self.createVariable('X', X.typecode(),
                                        dimensions=('X',), attrs=dict_)
                    # and set the data
                    self.variables['X'].data = N.array(X[:])
                        # note: X[:] is not a numpy array, but a older
                        # numeric array, thus I turn it into a numpy
                        # array.

                # add the variable
                for kk in input_file.variables.keys():
                    if not kk=='X':
                        var = input_file.variables[kk]
                        dict_ = self.fix_Scientific_bug(var.__dict__)
                        self.createVariable(kk, var.typecode(),
                                            dimensions=('X',),
                                            attrs=dict_)
                        # and set the data
                        self.variables[kk].data = N.array(var[:])

                # note that we just discarded the global attributs of
                # the input files
        # finally, close the config file
        config_file.close()

        ## make the time dimension and variable:
        time = N.arange(self.global_attrs['t_start'], self.global_attrs['t_final'],
                        self.global_attrs['t_output_dt'])
        self.createDimension('time', len(time))
        time_attrs = {'units': self.global_attrs['t_units'], 'standard_name': 'time'}
        time_var = self.createVariable('time', 'd', ('time',), time_attrs)
        time_var.data = time
        
        ## make the output variable
        thick_attrs = {'units': 'm', 'standard_name': 'land_ice_thickness'}
        self.createVariable('thick', 'd', ('X', 'time'), thick_attrs)

        # for the fortran module to know when ot output we make t_out_ind:
        # the indices of the time steps (dt) when to output
        self.time_steps_between_output = N.round(self.global_attrs['t_output_dt']/
                                            self.global_attrs['dt'])
        #: number of integeration steps
        self.t_steps = N.round(self.global_attrs['t_final']/
                                       self.global_attrs['dt'])
        
    def read_model_input(self):
        """Reads all the input files:
         - bed
         - initial thickness
         - mass balance distribution
        """
        pass

    def read_netCDF(self):
        """
        Reads a netCDF file (as saved by a previous model run) and
        puts everything into place, such that the model run can be
        contiued.
        """
        pass
    
    def write_netCDF(self):
        """
        Write all the calcualted data and also metadata to a netCDF
        file.
        """
        # open the file specified in the config file
        netCDFfile = helper_module.open_netCDF_into_python(
            self.global_attrs['output_filename'], 'w')
        # now follow U{http://gfesuite.noaa.gov/developer/netCDFPythonInterface.html}


        ## global attributs
        for kk in self.global_attrs:
            # this doggy stuff needs to be done because of Sientific
            # using Numeric arrays and not numpy arrays
            try:
                setattr(netCDFfile, kk, Numeric.array(self.global_attrs[kk]))
            except ValueError:
                setattr(netCDFfile, kk, str(self.global_attrs[kk]))

        
        ## first the dimensions
        for kk in self.dimensions:
            netCDFfile.createDimension(kk, self.dimensions[kk].size)
        ## now the variables
        for kk in self.variables:
            var = self.variables[kk]
            for kkk in var.attrs:
                # see comment above
                try:
                    setattr(netCDFfile, kkk, Numeric.array(var.attrs[kkk]))
                except ValueError:
                    setattr(netCDFfile, kkk, str(var.attrs[kkk]))
            netCDFvar = netCDFfile.createVariable(kk, var.datatype, var.dimensions)
            # fill in the data
            netCDFvar[:] = var.data
        ## write to disk
        netCDFfile.sync() 
    
    def run_model(self):
        """
        Runs the model according to the setup gathered from
        L{read_model_config} and L{read_input}.
        """
##thick_out = simple_ice_model(dt,t_steps,n_t_num,dx,rho,g_grav,a_rate,n_glen,xx,mass_b,bed,thick,thickr,thickl,[grid])

## in fortran
# subroutine simple_ice_model( dt, t_steps, n_t_num, dx, & ! numerics: 
# !  time step, total number of time steps, time steps between outputs, spatial discretisation step
#    rho, g_grav, A_rate, n_glen, & ! physical constants
#    xx, mass_b, bed, & ! input on grid: the grid, mass balance, bed elevation
#    thick, & !IC on grid: initial ice thickness
#    thickr, thickl, & ! BC: thickness at left and right
#    thick_out) ! output: the thickness at the specified time steps

        X = self.variables['X']
        land_ice_copy = copy.deepcopy(self.variables['land_ice_thickness'].data)
        dx = X.data[1]-X.data[0]
        time = self.variables['time']
        output_thickness = kees_simple_ice_model.simple_ice_model(
            self.global_attrs['dt'], self.t_steps, self.time_steps_between_output, dx,
            self.global_attrs['rho_i'], self.global_attrs['g_grav'], self.global_attrs['A_rate'], self.global_attrs['n_glen'],
            X.data, self.variables['land_ice_surface_specific_mass_balance'].data, self.variables['bedrock_altitude'].data,
            land_ice_copy,
            self.global_attrs['BC_thickl'], self.global_attrs['BC_thickr'], len(X.data))

        self.variables['thick'].data = output_thickness

    def plot_thickness(self, time):
        """
        Plot the ice thickness at a particular time on top of the bed.

        @type time: float
        @param time: The time when to plot the solution. If <0 then
                     the number is used as index (from the last).
        """
        # get what index is the solution at
        if time<0:
            ind = time
            time = self.variables['time'].data[-1]
        else:
            ind = N.where(time>self.variables['time'].data)[0][-1]

        fig = P.figure()
        P.plot(self.variables['X'].data,
               self.variables['bedrock_altitude'].data +
               self.variables['thick'].data[:,ind].T, 'b')
        P.hold(True)
        P.plot(self.variables['X'].data,
               self.variables['bedrock_altitude'].data, 'r',
               label=str(self.variables['bedrock_altitude'].attrs['standard_name']))
        P.xlabel(self.variables['time'].get_axis_label())
        P.ylabel('Elevation (%s)' % self.variables['bedrock_altitude'].attrs['units'])
        P.title('Solution at time %.0f (%s)'
                  %(time, self.variables['time'].attrs['units']))


    def plot_IC(self):
        """
        Plot the ice thickness at t0, together with the bed and mass
        balance distribution.
        """
        fig = P.figure()
        ax1 = P.subplot(2,1,1)
        P.plot(self.variables['X'].data,
               self.variables['land_ice_thickness'].data +
               self.variables['bedrock_altitude'].data , 'b',
               label=str(self.variables['land_ice_thickness'].attrs['standard_name']))
        P.hold(True)
        P.plot(self.variables['X'].data,
               self.variables['bedrock_altitude'].data, 'r',
               label=str(self.variables['bedrock_altitude'].attrs['standard_name']))
        P.xlabel(self.variables['X'].get_axis_label())
        P.ylabel('Elevation (%s)' % self.variables['bedrock_altitude'].attrs['units'])
        P.legend()
        P.title('IC and other input')

        ax1 = P.subplot(2,1,2)
        P.plot(self.variables['X'].data,
               self.variables['land_ice_surface_specific_mass_balance'].data)
        P.xlabel(self.variables['X'].get_axis_label())
        P.ylabel(self.variables['land_ice_surface_specific_mass_balance'].get_axis_label())



    # helper methods xxxxxxxxxxxxxxxxxxxxxxxxxxxxxx
    def createDimension(self, dimName, size):
        """
        Add a L{Dimension} instance.

        @type dimName: string
        @param dimName: The name of the dimension

        @type size: integer
        @param size: the length of the dimension
        """
        self.dimensions[dimName] = Dimension(dimName, size)
        return self.dimensions[dimName]

    def createVariable(self, varname, datatype='d',  dimensions=(), attrs={}):
        """
        Add a L{Dimension} instance.

        @type varname: string
        @param varname: The name of the variable

        @type dimensions: a touple of strings
        @param dimensions: what dimensions are associated with this
                           variable, if () then its scalar.
        
        @type attrs: dict
        @param attrs: a dictionary holding all the attirbutes of the
                      variable.
        """
        self.variables[varname] = Variable(varname, datatype
                                           ,  dimensions, attrs)
        return self.variables[varname]

    def fix_Scientific_bug(self, input_dict):
        """
        The C{Scientific.IO.NetCDF.NetCDFFile} retruns not numpy
        arrays but other one of a now outdated package numeric.  Thus
        all needs to be turned into numpy arrays.

        @type input_dict: dict
        @param input_dict: A dictionary containing numeric arrays

        @return: A dictionary containing numpy arrays
        """
        dict_ = {}
        for kk in input_dict:
            dict_[kk] = N.array(input_dict[kk])
        return dict_

    def get_svn_verison_number(self):
        """Return the version of the current svn revision"""

        status, output = commands.getstatusoutput("svn update")
        return output.split()[-1][:-1]

        

class Variable(object):
    """
    A class to hold variable data and attributes, similar to the
    C{Scientific.io.netcdf} variables.
    """

    def __init__(self,varname, datatype='d',  dimensions=(), attrs={}):
        """
        @type varname: string
        @param varname: Name of the variable.

        @type datatype: string
        @param datatype: One of:
                         - 'f4' (32-bit floating point)
                         - 'd' (64-bit floating point)
                         - 'i4' (32-bit signed integer)
                         - 'i2' (16-bit signed integer)
                         - 'i1' (8-bit signed integer)
                         - 'S1' (single-character string)

        @type dimensions: a touple of strings
        @param dimensions: what dimensions are associated with this
                           variable, if () then its scalar.
                           
        @type attrs: dict
        @param attrs: a dictionary holding all the attirbutes of the
                      variable.
        """
        #: the variable name
        self.name = varname
        #: the datatype
        self.datatype = datatype
        #: netCDF CF-1.4 attributes
        self.attrs = attrs

        self.dimensions = dimensions
        #: the acutal data in a numpy array (this could probably be done better)
        self.data = N.array([])

    def get_axis_label(self):
        """
        Returns a string to label a axis in a plot.
        """
        lab = '%s (%s)' %(self.attrs['standard_name'],
                           self.attrs['units'])
        return lab
        

class Dimension(object):
    """
    A class to hold a dimension.

    Note that each (netCDF) variable has a number of dimensions
    associated with it, i.e. a touple containing instances of this
    class.  The dimension should have a closely associated variable,
    I{coordinate variable} which needs to have the same name as the
    dimension.
    """

    def __init__(self, dimname, size):
        """
        @type dimname: string
        @param dimname: The standart name according CF 1.4 convention.

        @type size: int
        @param size: The length of the coordinate variable.
        """
        self.name = dimname
        self.size = size
